@@ -29,6 +29,8 @@
 #include "qwaylandtextinputinterface_p.h"
 #include "qwaylandinputcontext_p.h"
 #include "qwaylandinputmethodcontext_p.h"
+#include "qwaylandcallback_p.h"
+#include "qwaylandcursorsurface_p.h"
 
 #include <QtGui/private/qpixmap_raster_p.h>
 #include <QtGui/private/qguiapplication_p.h>
@@ -69,11 +71,12 @@ QWaylandInputDevice::Keyboard::Keyboard(QWaylandInputDevice *p)
             return;
         }
         mRepeatTimer.setInterval(1000 / mRepeatRate);
-        handleKey(mRepeatKey.time, QEvent::KeyRelease, mRepeatKey.key, mRepeatKey.modifiers,
-                  mRepeatKey.code, mRepeatKey.nativeVirtualKey, mRepeatKey.nativeModifiers,
+        Qt::KeyboardModifiers modifiers = this->modifiers();
+        handleKey(mRepeatKey.time, QEvent::KeyRelease, mRepeatKey.key, modifiers,
+                  mRepeatKey.code, mRepeatKey.nativeVirtualKey, this->mNativeModifiers,
                   mRepeatKey.text, true);
-        handleKey(mRepeatKey.time, QEvent::KeyPress, mRepeatKey.key, mRepeatKey.modifiers,
-                  mRepeatKey.code, mRepeatKey.nativeVirtualKey, mRepeatKey.nativeModifiers,
+        handleKey(mRepeatKey.time, QEvent::KeyPress, mRepeatKey.key, modifiers,
+                  mRepeatKey.code, mRepeatKey.nativeVirtualKey, this->mNativeModifiers,
                   mRepeatKey.text, true);
     });
 }
@@ -151,80 +154,6 @@ QWaylandWindow *QWaylandInputDevice::Pointer::focusWindow() const
 
 #if QT_CONFIG(cursor)
 
-class WlCallback : public QtWayland::wl_callback {
-public:
-    explicit WlCallback(::wl_callback *callback, std::function<void(uint32_t)> fn)
-        : QtWayland::wl_callback(callback)
-        , m_fn(fn)
-    {}
-    ~WlCallback() override { wl_callback_destroy(object()); }
-    void callback_done(uint32_t callback_data) override {
-        m_fn(callback_data);
-    }
-private:
-    std::function<void(uint32_t)> m_fn;
-};
-
-class CursorSurface : public QWaylandSurface
-{
-public:
-    explicit CursorSurface(QWaylandInputDevice::Pointer *pointer, QWaylandDisplay *display)
-        : QWaylandSurface(display)
-        , m_pointer(pointer)
-    {
-        connect(this, &QWaylandSurface::screensChanged,
-                m_pointer, &QWaylandInputDevice::Pointer::updateCursor);
-    }
-
-    void reset()
-    {
-        m_setSerial = 0;
-        m_hotspot = QPoint();
-    }
-
-    // Size and hotspot are in surface coordinates
-    void update(wl_buffer *buffer, const QPoint &hotspot, const QSize &size, int bufferScale, bool animated = false)
-    {
-        // Calling code needs to ensure buffer scale is supported if != 1
-        Q_ASSERT(bufferScale == 1 || version() >= 3);
-
-        auto enterSerial = m_pointer->mEnterSerial;
-        if (m_setSerial < enterSerial || m_hotspot != hotspot) {
-            m_pointer->set_cursor(m_pointer->mEnterSerial, object(), hotspot.x(), hotspot.y());
-            m_setSerial = enterSerial;
-            m_hotspot = hotspot;
-        }
-
-        if (version() >= 3)
-            set_buffer_scale(bufferScale);
-
-        attach(buffer, 0, 0);
-        damage(0, 0, size.width(), size.height());
-        m_frameCallback.reset();
-        if (animated) {
-            m_frameCallback.reset(new WlCallback(frame(), [this](uint32_t time){
-               Q_UNUSED(time);
-               m_pointer->cursorFrameCallback();
-            }));
-        }
-        commit();
-    }
-
-    int outputScale() const
-    {
-        int scale = 0;
-        for (auto *screen : m_screens)
-            scale = qMax(scale, screen->scale());
-        return scale;
-    }
-
-private:
-    QScopedPointer<WlCallback> m_frameCallback;
-    QWaylandInputDevice::Pointer *m_pointer = nullptr;
-    uint m_setSerial = 0;
-    QPoint m_hotspot;
-};
-
 int QWaylandInputDevice::Pointer::idealCursorScale() const
 {
     if (seat()->mQDisplay->compositor()->version() < 3) {
@@ -341,7 +270,8 @@ void QWaylandInputDevice::Pointer::updateCursor()
     qCWarning(lcQpaWayland) << "Unable to change to cursor" << shape;
 }
 
-CursorSurface *QWaylandInputDevice::Pointer::getOrCreateCursorSurface()
+CursorSurface<QWaylandInputDevice::Pointer> *
+QWaylandInputDevice::Pointer::getOrCreateCursorSurface()
 {
     if (!mCursor.surface)
         mCursor.surface.reset(new CursorSurface(this, seat()->mQDisplay));
@@ -384,6 +314,7 @@ QWaylandInputDevice::QWaylandInputDevice(QWaylandDisplay *display, int version,
     : QtWayland::wl_seat(display->wl_registry(), id, qMin(version, 9))
     , mQDisplay(display)
     , mDisplay(display->wl_display())
+    , mId(id)
 {
 #if QT_CONFIG(wayland_datadevice)
     if (mQDisplay->dndSelectionHandler()) {
@@ -418,8 +349,13 @@ QWaylandInputDevice::QWaylandInputDevice(QWaylandDisplay *display, int version,
 #endif
 }
 
-// Can't be in header because dtors for scoped pointers aren't known there.
-QWaylandInputDevice::~QWaylandInputDevice() = default;
+QWaylandInputDevice::~QWaylandInputDevice()
+{
+    if (version() >= WL_SEAT_RELEASE_SINCE_VERSION)
+        release();
+    else
+        wl_seat_destroy(object());
+}
 
 void QWaylandInputDevice::seat_capabilities(uint32_t caps)
 {
@@ -440,7 +376,7 @@ void QWaylandInputDevice::seat_capabilities(uint32_t caps)
             mTouchPadDevice = new QPointingDevice(
                         QLatin1String("touchpad"), 0, QInputDevice::DeviceType::TouchPad,
                         QPointingDevice::PointerType::Finger, QInputDevice::Capability::Position,
-                        MaxTouchPoints, 0, QString(), QPointingDeviceUniqueId(), this);
+                        MaxTouchPoints, 0, mSeatName, QPointingDeviceUniqueId(), this);
             QWindowSystemInterface::registerInputDevice(mTouchPadDevice);
             mPointerGesturePinch.reset(pointerGestures->createPointerGesturePinch(this));
             mPointerGesturePinch->init(pointerGestures->get_pinch_gesture(mPointer->object()));
@@ -461,7 +397,7 @@ void QWaylandInputDevice::seat_capabilities(uint32_t caps)
             mTouchDevice = new QPointingDevice(
                         QLatin1String("some touchscreen"), 0, QInputDevice::DeviceType::TouchScreen,
                         QPointingDevice::PointerType::Finger, QInputDevice::Capability::Position,
-                        MaxTouchPoints, 0,QString(), QPointingDeviceUniqueId(), this);
+                        MaxTouchPoints, 0, mSeatName, QPointingDeviceUniqueId(), this);
             QWindowSystemInterface::registerInputDevice(mTouchDevice);
         }
     } else if (!(caps & WL_SEAT_CAPABILITY_TOUCH) && mTouch) {
@@ -522,6 +458,12 @@ void QWaylandInputDevice::handleEndDrag()
         mPointer->releaseButtons();
 }
 
+void QWaylandInputDevice::handleStartDrag()
+{
+    if (mPointer)
+        mPointer->leavePointers();
+}
+
 #if QT_CONFIG(wayland_datadevice)
 void QWaylandInputDevice::setDataDevice(QWaylandDataDevice *device)
 {
@@ -551,6 +493,18 @@ void QWaylandInputDevice::setTextInput(QWaylandTextInputInterface *textInput)
     mTextInput.reset(textInput);
 }
 
+#if QT_CONFIG(tabletevent)
+void QWaylandInputDevice::setTabletSeat(QWaylandTabletSeatV2 *tabletSeat)
+{
+    mTabletSeat.reset(tabletSeat);
+}
+
+QWaylandTabletSeatV2 *QWaylandInputDevice::tabletSeat() const
+{
+    return mTabletSeat.get();
+}
+#endif
+
 void QWaylandInputDevice::setTextInputMethod(QWaylandTextInputMethod *textInputMethod)
 {
     mTextInputMethod.reset(textInputMethod);
@@ -654,6 +608,9 @@ void QWaylandInputDevice::setCursor(const QCursor *cursor, const QSharedPointer<
 
     if (mPointer)
         mPointer->updateCursor();
+
+    if (mTabletSeat)
+        mTabletSeat->updateCursor();
 }
 #endif
 
@@ -829,6 +786,8 @@ void QWaylandInputDevice::Pointer::pointer_button(uint32_t serial, uint32_t time
     default: return; // invalid button number (as far as Qt is concerned)
     }
 
+    mLastButton = qt_button;
+
     if (state)
         mButtons |= qt_button;
     else
@@ -867,10 +826,13 @@ void QWaylandInputDevice::Pointer::invalidateFocus()
 
 void QWaylandInputDevice::Pointer::releaseButtons()
 {
-    mButtons = Qt::NoButton;
+    setFrameEvent(new ReleaseEvent(nullptr, mParent->mTime, mSurfacePos, mGlobalPos, Qt::NoButton, Qt::NoButton, mParent->modifiers()));
+}
 
+void QWaylandInputDevice::Pointer::leavePointers()
+{
     if (auto *window = focusWindow()) {
-        ReleaseEvent e(focusWindow(), mParent->mTime, mSurfacePos, mGlobalPos, mButtons, Qt::NoButton, mParent->modifiers());
+        LeaveEvent e(focusWindow(), mSurfacePos, mGlobalPos);
         window->handleMouse(mParent, e);
     }
 }
@@ -881,8 +843,10 @@ public:
     WheelEvent(QWaylandWindow *surface, Qt::ScrollPhase phase, ulong timestamp, const QPointF &local,
                const QPointF &global, const QPoint &pixelDelta, const QPoint &angleDelta,
                Qt::MouseEventSource source, Qt::KeyboardModifiers modifiers, bool inverted)
-        : QWaylandPointerEvent(QEvent::Wheel, phase, surface, timestamp,
-                               local, global, pixelDelta, angleDelta, source, modifiers, inverted)
+        : QWaylandPointerEvent(QEvent::Wheel, phase, surface, timestamp, local, global,
+                               modifiers & Qt::AltModifier ? pixelDelta.transposed() : pixelDelta,
+                               modifiers & Qt::AltModifier ? angleDelta.transposed() : angleDelta,
+                               source, modifiers, inverted)
     {
     }
 };
@@ -1175,10 +1139,12 @@ void QWaylandInputDevice::Pointer::flushFrameEvent()
         } else if (mFrameData.event->type == QEvent::MouseButtonRelease) {
             // If the window has been destroyed, we still need to report an up event, but it can't
             // be handled by the destroyed window (obviously), so send the event here instead.
-            QWindowSystemInterface::handleMouseEvent(nullptr, event->timestamp, event->local,
-                                 event->global, event->buttons,
-                                 event->button, event->type,
-                                 event->modifiers);// , Qt::MouseEventSource source = Qt::MouseEventNotSynthesized);
+            QWindowSystemInterface::handleMouseEvent(
+                    nullptr, event->timestamp,
+                    QPointingDevice::primaryPointingDevice(mParent->seatname()), event->local,
+                    event->global, event->buttons, event->button, event->type,
+                    event->modifiers); // , Qt::MouseEventSource source =
+                                       // Qt::MouseEventNotSynthesized);
         }
         delete mFrameData.event;
         mFrameData.event = nullptr;
@@ -1349,8 +1315,6 @@ void QWaylandInputDevice::Keyboard::keyboard_key(uint32_t serial, uint32_t time,
             mRepeatKey.code = code;
             mRepeatKey.time = time;
             mRepeatKey.text = text;
-            mRepeatKey.modifiers = modifiers;
-            mRepeatKey.nativeModifiers = mNativeModifiers;
             mRepeatKey.nativeVirtualKey = sym;
             mRepeatTimer.setInterval(mRepeatDelay);
             mRepeatTimer.start();
@@ -1551,9 +1515,9 @@ void QWaylandInputDevice::Touch::touch_frame()
         const QWindowSystemInterface::TouchPoint &tp = mPendingTouchPoints.constLast();
         // When the touch event is received, the global pos is calculated with the margins
         // in mind. Now we need to adjust again to get the correct local pos back.
-        QMargins margins = window->frameMargins();
+        QMargins margins = mFocus->clientSideMargins();
         QPoint p = tp.area.center().toPoint();
-        QPointF localPos(window->mapFromGlobal(QPoint(p.x() + margins.left(), p.y() + margins.top())));
+        QPointF localPos(mFocus->mapFromGlobal(p) + QPoint(margins.left(), margins.top()));
         if (mFocus->touchDragDecoration(mParent, localPos, tp.area.center(), tp.state, mParent->modifiers()))
             return;
     }
